// ORIGINAL COPYRIGHT
// cfdisk was inspired by the fdisk program, by A. V. Le Blanc (LeBlanc@mcc.ac.uk).
// Copyright (C) 1994 Kevin E. Martin (martin@cs.unc.edu)
// Created:	Fri Jan 28 22:46:58 1994, martin@cs.unc.edu
// >2GB patches: Sat Feb 11 09:08:10 1995, faith@cs.unc.edu
// Prettier menus: Sat Feb 11 09:08:25 1995, Janne Kukonlehto <jtklehto@stekt.oulu.fi>
// Versions 0.8e-g: aeb@cwi.nl

#include "config.h"

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <math.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#ifndef __FreeBSD__
#ifdef HAVE_DIETLIBC
#include <asm/types.h>
#else
#include <getopt.h>
#include <linux/genhd.h>
#include <linux/hdreg.h>
#include <linux/fs.h>
#endif
#endif
#include <badpenguin-chain.h>
#include <badpenguin-errors.h>
#include "badpenguin-hardware.h"
#include "set-proc-dir.h"

// from linux kernel
#define REISERFS_SUPER_MAGIC_STRING "ReIsErFs"
#define REISER2FS_SUPER_MAGIC_STRING "ReIsEr2Fs"

#define mybzero(X,Y)	memset( (X), 0, (Y) * sizeof(char) )

#define SECTOR_SIZE 	512
#define PARTITION_SIZE 	16
#define SECTOR_MAGIC	0xAA55

#pragma pack(1)
struct stPartitionTable {
  __u8 bootFlag;	/* 0x80 - active */
  __u8 head1;		/* starting head */
  __u8 sector1;		/* starting sector */
  __u8 cylinder1;	/* starting cylinder */
  __u8 type;		/* What partition type */
  __u8 head2;		/* end head */
  __u8 sector2;		/* end sector */
  __u8 cylinder2;	/* end cylinder */
 __u32 offset;		/* starting sector counting from 0 */
 __u32 size;		/* nr of sectors in partition */
}  ;

struct stBootSector {
  __u8 buffer[0x1be];
  struct stPartitionTable partition[4];
  __u16 magic;
};
#pragma pack()



int read_sector(int fd, struct stBootSector *buffer, int sector, char *device) {
	int ecount=0;
#ifdef HAVE_DIETLIBC
  loff_t offset = ((loff_t)sector) * SECTOR_SIZE;
read1:
  if ( lseek64(fd, (loff_t)offset, 0) < 0 ) {
#else
read1:
  if ( llseek(fd, ((loff_t) sector)*SECTOR_SIZE, SEEK_SET) < 0 ) {
#endif
    if (ecount<5) {
      ecount++;
      sleep(1);
      goto read1;
    }
    fprintf(stderr, "libcfdisk: unable to open sector %i on device %s \n", sector, device);
    return(ERROR_IO);
  }
read2:
  if ( SECTOR_SIZE != read(fd, buffer, SECTOR_SIZE) ) {
    if (ecount<5) {
      ecount++;
      sleep(1);
      goto read2;
    }
    fprintf(stderr, "libcfdisk: error reading sector %i on device %s \n", sector, device );  
    return(ERROR_IO);
  }
  return 0;
}


TPartition *create_partition() {
  return calloc( 1, sizeof(TPartition));
}


char *decode_filesystem(int type) {
  switch (type) {
    case 0x00: return strdup("Empty");
    case 0x01: return strdup("DOS 12-bit FAT");
    case 0x02: return strdup("XENIX /");
    case 0x03: return strdup("XENIX /usr");
    case 0x04: return strdup("DOS 16-bit FAT <32M");
    case 0x05: return strdup("Extended (contains other slice)");
    case 0x06: return strdup("DOS 16-bit FAT >=32M");
    case 0x07: return strdup("OS/2 HPFS or NTFS");
    case 0x08: return strdup("AIX boot");
    case 0x09: return strdup("AIX data");
    case 0x0A: return strdup("OS/2 Boot Manager");
    case 0x0B: return strdup("Win95 FAT32");
    case 0x0C: return strdup("Win95 FAT32 (LBA)");
    case 0x0E: return strdup("Win95 FAT16 (LBA)");
    case 0x0F: return strdup("Win98 Extended (LBA)");
    case 0x11: return strdup("Hidden DOS FAT12");
    case 0x12: return strdup("Compaq diagnostics");
    case 0x14: return strdup("Hidden DOS FAT16");
    case 0x16: return strdup("Hidden DOS FAT16 (big)");
    case 0x17: return strdup("Hidden HPFS / NTFS");
    case 0x63: return strdup("GNU HURD");
    case 0x77: return strdup("QNX 4.x 1th part.");
    case 0x78: return strdup("QNX 4.x 2nd part.");
    case 0x79: return strdup("QNX 4.x 3th part.");
    case 0x80: return strdup("Old Minix");
    case 0x81: return strdup("Linux/Minix");

    case 0x82: return strdup("Linux Swap");
    case 0x83: return strdup("Linux Ext2Fs");
    case 0x85: return strdup("Linux Extended");

    case 0xA0: return strdup("Notebook hibernation area");
    case 0xA5: return strdup("BSD/386");

    case 0xFD: return strdup("Linux Raid");

  }
  return strdup("Unknown");
}


char *decode_module(int type) {
  switch (type) {
    case 0x01:
    case 0x04:
    case 0x06:
	 return strdup(FILESYSTEM_MSDOS);
    case 0x0B:
    case 0x0C:
    case 0x0E:
/*
    gm@badpenguin.org reported that this is an EXTENDED partition!
    case 0x0F:
*/
	return strdup(FILESYSTEM_VFAT);
    case 0x82:
	return strdup(FILESYSTEM_SWAP);
    case 0x83:
	return strdup(FILESYSTEM_EXT2);
  }
  return strdup(FILESYSTEM_NONE);
}

int magic_detect_ext2(char* device, int partno) {
  char subdev[80];
  int fd2;
  char magic[256];
//  int jm;
  // partno is passed from 0 to 4 we have to increase it when making the device
  snprintf( subdev, 80, "%s%i", device, partno );
  //printf("probing %s\n", subdev);
  fd2 = open(subdev, O_RDONLY);
  //printf("opened %i\n", fd2);
  lseek(fd2, 1024, SEEK_SET);
  mybzero(&magic, 256);
  //printf("read=%i\n", read(fd2, &magic, 256 ) );
  read(fd2, &magic, 256);
  //for (jm=56; jm<=57; jm++) printf("%2X", (unsigned char)magic[jm] );
  if ( 
    ((unsigned char)magic[56] == 0x53) &&
    ((unsigned char)magic[57] == 0xEF)
  ) {
    //printf("EXT2 DETECTED!");
    close(fd2);
    return 0;
  }

  //=========== EXT2 FAILED TRY WITH REISER =============
  lseek(fd2, 65536, SEEK_SET);
  mybzero(&magic, 256);
  read(fd2, &magic, 256);
  //UNUSED: magic[52+13]=0;
//  printf("%s:%12s< \n", subdev, magic+52 );
  close(fd2);

  if ( 0 == strncmp( magic+52, 
    REISER2FS_SUPER_MAGIC_STRING, strlen(REISER2FS_SUPER_MAGIC_STRING) 
  ) ) {
    return 1;
  }
  if ( 0 == strncmp( magic+52, 
    REISERFS_SUPER_MAGIC_STRING, strlen(REISERFS_SUPER_MAGIC_STRING) 
  ) ) {
    return 1;
  }

  return -1;
}



int probe_partition(TChain *part, char *device) {
  int fd, i;
  struct stBootSector lpstBootSector, lpstExtSector;
  long long extstart, extsize;
  int extloop;
  int extnum;
  int ptype;

  TChainItem *pitem;
  TPartition *ppart;
  
//  assert( sizeof(struct stPartitionTable) == PARTITION_SIZE );
//  assert( sizeof(struct stBootSector) == SECTOR_SIZE );

  // Open Disk
  fd = open(device, O_RDONLY);
  if (fd<=0) {
    fprintf(stderr, "libcfdisk: unable to open %s \n", device);
    return(ERROR_IO);
  }
  // Read Sector
  mybzero( &lpstBootSector, sizeof(lpstBootSector) );
  if ( 0 != read_sector(fd, &lpstBootSector, 0, device)) {
    return 1;
  }
  // Parse boot record
  if ( lpstBootSector.magic != SECTOR_MAGIC ) {
    printf( "libcfdisk: unknow magic numer %X on %s\n" , lpstBootSector.magic, device );
    return 1;
  }

  extnum=5;
  for (i = 0; i < 4; i++) {
  
    // IGNORE EMPTY PARTITIONS
    if ( lpstBootSector.partition[i].type == 0 ) continue;

    /* Attach new partition */
    pitem = chain_append(part);
    ppart = create_partition();
    pitem->data = ppart;
    /* fill data */
    ppart->device = calloc( 12, sizeof(char) );
    snprintf( ppart->device, 12, "%s%i", device, i+1 );
    ppart->type = lpstBootSector.partition[i].type;
    // AGX
    if (ppart->type==0x83) {
      ptype = magic_detect_ext2( device, i+1);
      if (ptype == 0) {
        ppart->filesystem = decode_filesystem(ppart->type);
        ppart->module     = decode_module(ppart->type);
      } else
      if (ptype == 1) {
        ppart->filesystem = strdup("Linux ReiserFS");
        ppart->module     = strdup(FILESYSTEM_REISER);
      } else {
        ppart->filesystem = strdup("Linux Unknown");
        ppart->module     = decode_module(ppart->type);
      }
    } else {
      ppart->filesystem = decode_filesystem(ppart->type);
      ppart->module     = decode_module(ppart->type);
    }
    ppart->bootflag = lpstBootSector.partition[i].bootFlag;
    ppart->size     = lpstBootSector.partition[i].size;
    
    if ( 
      (lpstBootSector.partition[i].type == 0x05) || 
      (lpstBootSector.partition[i].type == 0x0F) || 
      (lpstBootSector.partition[i].type == 0x85)
    ) {
      extloop=0;
      extstart=lpstBootSector.partition[i].offset;
      extsize=0;
      while (extloop==0) {
        mybzero( &lpstExtSector, sizeof(lpstExtSector) );
        //printf( "> %u \n" ,  extstart+extsize );
        if (0 != read_sector(fd, &lpstExtSector, extstart+extsize, device )) {
		break;
	}

        /* Attach new partition */
        pitem = chain_append(part);
        ppart = create_partition();
        pitem->data = ppart;
        /* fill data */
        ppart->device = calloc( 12, sizeof(char) );
        snprintf( ppart->device, 12, "%s%i", device, extnum );
        ppart->type = lpstExtSector.partition[0].type;
//        ppart->filesystem = decode_filesystem(ppart->type);
//	ppart->module     = decode_module(ppart->type);
        ppart->bootflag = lpstExtSector.partition[0].bootFlag;
        ppart->size     = lpstExtSector.partition[0].size;

      // AGX (we use extnum here!!!)
    if (ppart->type==0x83) {
      ptype = magic_detect_ext2( device, extnum);
      if (ptype == 0) {
        ppart->filesystem = decode_filesystem(ppart->type);
        ppart->module     = decode_module(ppart->type);
      } else
      if (ptype == 1) {
        ppart->filesystem = strdup("Linux ReiserFS");
        ppart->module     = strdup(FILESYSTEM_REISER);
      } else {
        ppart->filesystem = strdup("Linux Unknown");
        ppart->module     = decode_module(ppart->type);
      }
    } else {
      ppart->filesystem = decode_filesystem(ppart->type);
      ppart->module     = decode_module(ppart->type);
    }

        // next
	extsize = lpstExtSector.partition[1].offset;
	extnum++;

        if ( 
	  (lpstExtSector.partition[1].type != 0x05) && 
	  (lpstExtSector.partition[1].type != 0x0F) && 
	  (lpstExtSector.partition[1].type != 0x85)
	) {
	  extloop = 1;
	}

      }
    }
  }
  // Close Disk
  if ( 0 != close(fd) ) {
    fprintf(stderr, "libcfdisk: unable to close %s \n", device);
    return(ERROR_IO);
  }
  return 0;  
}


int add_cdrom( TChain *part, TBlockDevice *device ) {
  TChainItem *pitem;
  TPartition *ppart;
  
    /* Attach new partition */
    pitem = chain_append(part);
    ppart = create_partition();
    pitem->data = ppart;
    /* fill data */
    ppart->device = strdup(device->device);
    ppart->type = 0xFF;
    ppart->filesystem = strdup(device->model);
    ppart->module     = strdup(FILESYSTEM_ISO9660);
    ppart->bootflag = 0;
    ppart->size     = 0;
  return 0;
}


TChain *query_partitions(TChain *part, TChain *eide, TChain *scsi) {
  TChainItem *c;
  TBlockDevice *d;

  if ( initProc == 0 ) {
    set_proc_dir(DEFAULT_PROC_DIR, DEFAULT_DETECT_DIR );
  }
  
  if (part!=NULL) {
    chain_cleanall(part);
  } else {
    part = chain_create();
  }
  
  c = eide->head;
  while (c != NULL ) {
    d = TBLOCKDEVICE(c->data);
    if ( 0 == strcmp(d->media,MEDIA_CDROM_EIDE) ) {
      add_cdrom( part, d );
    } else
    if ( 0 == strcmp(d->media,MEDIA_DISK_EIDE) ) {
      probe_partition( part, d->device );
    }
    c = c -> next;
  }

  c = scsi->head;
  while (c != NULL ) {
    d = TBLOCKDEVICE(c->data);
    if ( 0 == strcmp(d->media,MEDIA_CDROM_SCSI) ) {
      add_cdrom( part, d );
    } else
    if ( 0 == strcmp(d->media,MEDIA_DISK_SCSI) ) {
      probe_partition( part, d->device );
    }
    c = c -> next;
  }
  
  return part;
}

